/* -----------------------------------------------------------------------------
 * d.cxx
 *
 *     D wrapper module.
 *
 * Author(s) : Andy Friesen
 *
 * Copyright (C) 2004
 * See the file LICENSE for information on usage and redistribution.
 * ----------------------------------------------------------------------------- */

char cvsroot_d_cxx[] = "$Header: /cvsroot/SWIG/Source/Modules/d.cxx,v 1.50 2003/12/23 21:45:38 cheetah Exp $";

#include <limits.h> // for INT_MAX
#include <ctype.h>
#include "swigmod.h"
#ifndef MACSWIG
#   include "swigconfig.h"
#endif

#include <cassert>
#include <string>
#include <vector>
#include <map>

namespace {

    template <typename T>
    void destroy(T* t) {
        Delete(t);
    }

    template<>
    void destroy<Wrapper>(Wrapper* w) {
        DelWrapper(w);
    }

    struct ScopedPtrBase { };

    template <class T>
    struct ScopedPtr : ScopedPtrBase {
        ScopedPtr(T* t)  : _data(t) { }
        ScopedPtr()      : _data(0) { }

        ScopedPtr(ScopedPtr& rhs) {
            (*this) = rhs;
        }

        virtual ~ScopedPtr() { 
            destroy(_data);
        }

        T* operator ->() const { return  _data; }
        T* get()         const { return  _data; }
     
        ScopedPtr& operator = (T* t) {
            destroy(_data);
            _data = t;
            return *this;
        }

        ScopedPtr& operator = (ScopedPtr& rhs) {
            (*this) = rhs.get();
            rhs = 0;
            return *this;
        }

    private:
        T* _data;
        //ScopedPtr(ScopedPtr&) { } // disallow copying
    };

    typedef ScopedPtr<File> ScopedFile;
    typedef ScopedPtr<String> ScopedString;

    template <typename T>
    ScopedPtr<T> makeAutoPtr(T* t) {
        return ScopedPtr<T>(t);
    }

    struct DHash {
        DHash(DOHHash* p)
            : _hash(p)
            , _own(false)
        {}

        ~DHash() {
            if (_own) {
                destroy(_hash);
            }
        }

        DOH* getattr(DOH* name) const {
            return Getattr(_hash, name);
        }

        void setattr(DOH* name, DOH* value) {
            Setattr(_hash, name, value);
        }

        struct ValueProxy {
            friend struct DHash;

            operator DOH*() const {
                return _value;
            }

            ValueProxy& operator = (DOH* rhs) {
                _this->setattr(_key, rhs);
                return *this;
            }

            ValueProxy(const ValueProxy& rhs)
                : _this(rhs._this)
                , _key(rhs._key)
                , _value(rhs._value)
            { }

        private:
            ValueProxy(DHash* parent, DOH* key)
                : _this(parent)
                , _key(key)
                , _value(Getattr(parent, key))
            { }

            DHash* _this;
            DOH* _key;
            DOH* _value;
        };


        DOH* operator[](DOH* key) const {
            return getattr(key);
        }

        ValueProxy operator[](DOH* key) {
            return ValueProxy(this, key);
        }

    private:
        DOHHash* _hash;
        bool _own;
    };
}

// Evil little macro for automatically destroying a local when its scope dies.
// magic: Assigning a temporary to a const& extends the life of the temporary
// to that of the reference.
#define AUTO_HIDDEN(x, y) x ## y
#define AUTO_HIDE2(x, y) AUTO_HIDDEN(x, y)
#define AUTO(x) ScopedPtrBase const& AUTO_HIDE2(auto_secret, __LINE__) = makeAutoPtr(x);

class D : public Language {
    /*
     * TODO: deal with nested types.
     */

    virtual void main(int argc, char** argv) {
        SWIG_library_directory("d");
        Preprocessor_define("SWIGD 1", 0);
        SWIG_config_file("d.swg");
        SWIG_typemap_lang("d");

        ScopedString cname = NewStringf("test.cpp");
        ScopedString dimp  = NewStringf("test_extern.d");
        ScopedString dname = NewStringf("test.d");

        c_source = NewFile(cname.get(), "w");
        d_import = NewFile(dimp .get(), "w");
        d_source = NewFile(dname.get(), "w");

#if 0
        c_init   = NewString("");
#endif

        // SWIG uses these names to insert code in various places.
        // Link them up with actual file handles.
        Swig_register_filebyname("runtime", c_source.get());
        Swig_register_filebyname("wrapper", d_source.get());

        // We don't actually have a C header: 
        // just put it in the main C++ source file for now.
        Swig_register_filebyname("header", c_source.get());

        allow_overloading();
    }

    virtual int top(Node* n) {
        Language::top(n);
        return SWIG_OK;
    }

    //
    virtual int enumDeclaration(Node* n) {
        Printf(stderr, "enum %s\n", Getattr(n, "name"));

        Printf(d_source.get(), "enum %s {\n", Getattr(n, "name"));
        Language::enumDeclaration(n);
        Printf(d_source.get(), "}\n");
        return SWIG_OK;
    }

    virtual int enumvalueDeclaration(Node* n) {
        Printf(d_source.get(), "    %s = %s,\n", Getattr(n, "name"), Getattr(n, "enumvalue"));
        return SWIG_OK;
    }

    //
    virtual int globalfunctionHandler(Node* n) {
        if (Getattr(n, "sym:overloaded") != 0) {
            // overloads need proxies so overloading works correctly.
            // pass-through to generate proxies.
            Printf(stderr, "global overloaded function %s\n", Getattr(n, "name"));

            return Language::globalfunctionHandler(n);
        } else {
            Printf(stderr, "Skipping non-overloaded global function %s: D needs no glue.\n", Getattr(n, "name"));
            return SWIG_NOWRAP;
        }
    }

    virtual int functionHandler(Node* node) {
        if (Getattr(node, "sym:overloaded") != 0) {
            // Overloads aren't jiggy with extern "C".  Need proxies.
            // add wart to name to make it unique.
            String* newName = NewString("");
            Printf(newName, "%s%s", Getattr(node, "sym:name"), Getattr(node, "sym:overname"));
            Setattr(node, "sym:name", newName);
        }
        return Language::functionHandler(node);
    }

    virtual int memberfunctionHandler(Node* n) {
        Printf(stderr, "\tmethod: %s\n", Getattr(n, "name"));

        return Language::memberfunctionHandler(n);
    }

    virtual int staticmemberfunctionHandler(Node* n) {
        Printf(stderr, "\tstatic method %s\n", Getattr(n, "name"));
        return Language::staticmemberfunctionHandler(n);
    }

    virtual int callbackfunctionHandler(Node* n) {
        return NYI(n);
    }

    //
    virtual int globalvariableHandler(Node *n) {
        Printf(stderr, "\tglobal attribute %s\n", Getattr(n, "name"));
        return Language::globalvariableHandler(n);
    }

    virtual int membervariableHandler(Node *n) {
        Printf(stderr, "\tattribute %s\n", Getattr(n, "name"));
        return Language::membervariableHandler(n);
    }

    virtual int staticmembervariableHandler(Node *n) {
        Printf(stderr, "\tstatic attribute %s\n", Getattr(n, "name"));
        return Language::staticmembervariableHandler(n);
    }

    //
    virtual int memberconstantHandler(Node* n) {
        Printf(stderr, "\tconstant attribute %s\n", Getattr(n, "name"));
        return Language::memberconstantHandler(n);
    }

    virtual int constructorHandler(Node* n) {
        Printf(stderr, "\tctor %s\n", Getattr(n, "name"));
        return Language::constructorHandler(n);
    }

    virtual int copyconstructorHandler(Node* n) {
        Printf(stderr, "\tcopy ctor %s\n", Getattr(n, "name"));                
        return Language::copyconstructorHandler(n);
    }

    virtual int destructorHandler(Node* n) {
        Printf(stderr, "\tdtor\n");
        return Language::destructorHandler(n);
    }

    virtual int classHandler(Node* n) {
        Printf(stderr, "\nclass %s:\n", Getattr(n, "name"));

        String* className = Getattr(n, "name");

        Language::classHandler(n);

        return SWIG_OK;
    }

    //
    virtual int constantWrapper(Node* n) {
        dumpNode(n);
        String* varType = Getattr(n, "type");
        String* varName = Getattr(n, "name");
        String* varValue = Getattr(n, "value");

        int valueType = SwigType_type(varType);
        if (valueType == T_STRING || valueType == T_CHAR) {
            // put the quotes back (SWIG strips them for some reason I cannot fathom)
            char quote = valueType == T_STRING ? '\"' : '\'';
            String* oldValue = varValue;

            varValue = NewString("");

            Printf(varValue, "%c%s%c", quote, oldValue, quote);
            Setattr(n, "value", varValue);
        }

        Printf(d_source.get(), "const %s %s = %s;\n", varType, varName, varValue);

        return SWIG_OK;
    }

    virtual int functionWrapper(Node* node) {
        int result;

        // C glue
        result = writeExternCFunc(node);
        if (result != SWIG_OK) {
            return result;
        }

        // D glue
        result = writeExternDFunc(node);
        if (result != SWIG_OK) {
            return result;
        }

        // Now the D wrapper
        bool isMethod = false;
        Node* parent = Getattr(node, "parentNode");
        if (Cmp(Getattr(parent, "nodeType"), "class") == 0) {
            dumpNode(node);
        }

        if (1 || Cmp(Getattr(parent, "nodeType"), "class") == 0) {
            Printf(stderr, "TODO: wrap member functions\n");
        } else {
            // global scope?
            result = writeWrapperFunc(node);
        }

        return result;
    }

    int writeExternCFunc(Node* node) {
        String*    name  = Getattr(node,"sym:name");
        SwigType*  type  = Getattr(node,"type");
        ParmList*  parms = Getattr(node,"parms");

        String* returnType = SwigType_str(type, 0);
        AUTO(returnType);

        Wrapper* wrapper = NewWrapper();
        AUTO(wrapper);

        emit_attach_parmmaps(parms, wrapper);
        Swig_typemap_attach_parms("CtoExtern", parms, wrapper);

        Printf(wrapper->def, "DllExport %s CDecl %s(%s) {", returnType, name, ParmList_str(parms));

        emit_args(type, parms, wrapper);

        emit_action(node, wrapper);

        if (SwigType_type(type) != T_VOID) {
            Printf(wrapper->code, "return result;");
        }

        Printf(wrapper->code, "}");
        Wrapper_print(wrapper, c_source.get());

        writeExternDFunc(node);

        return SWIG_OK;
    }

    int writeExternDFunc(Node* node) {
        String*    name  = Getattr(node,"sym:name");
        SwigType*  type  = Getattr(node,"type");
        ParmList*  parms = Getattr(node,"parms");

        String*    argList = NewString("");
        AUTO(argList);

        String* returnType = SwigType_str(type, 0);
        AUTO(returnType);

        Swig_typemap_attach_parms("CtoD", parms, 0);

        String* mappedReturnType = Swig_typemap_lookup_new("CtoD", node, "", 0);

        int parmCount = emit_num_arguments(parms);

        // Construct argument list: transform each argument type
        // and make up a name.
        Parm* parm = parms;
        for (int i = 0; i < parmCount; i++) {
            String* mappedType = Getattr(parm, "tmap:CtoD");

            String* oldType = SwigType_str(Getattr(parm, "type"), 0);

            if (mappedType != 0) {
                Printv(argList, mappedType, NIL);
            } else {
                Printf(stderr, "Can't convert argument type %s in function %s\n", oldType, name);
                Printv(argList, oldType, NIL);
            }

            Printf(argList, " arg%i", i + 1);

            if (i < parmCount - 1) {
                Printf(argList, ", ");
            }

            parm = nextSibling(parm);
        }

        if (mappedReturnType == 0) {
            mappedReturnType = Copy(returnType);
        }

        Printf(d_import.get(), "extern (C) %s %s(%s);\n", mappedReturnType, name, argList);

        return SWIG_OK;
    }

    int writeWrapperFunc(Node* node) {
        dumpNode(node);

        DHash hash = node;

        /*String*   name  = Getattr(node, "name");        // pretty name
        String*   wrapName = Getattr(node, "sym:name"); // mangled name
        SwigType* type  = Getattr(node, "type");        // return type
        ParmList* parms = Getattr(node, "parms");*/
        String* name = hash["name"];
        String* wrapName = hash["sym:name"];
        SwigType* type = hash["type"];
        ParmList* parms = hash["parms"];

        Wrapper* wrapper = NewWrapper();
        AUTO(wrapper);

        Swig_typemap_attach_parms("DtoC", parms, wrapper);

        String* argList = NewString("");
        AUTO(argList);

        Parm* parm = parms;
        int parmCount = emit_num_arguments(parms);
        for (int i = 0; i < parmCount; i++) {
            String* mappedType = Getattr(parm, "tmap:CtoD");

            String* oldType = SwigType_str(Getattr(parm, "type"), 0);

            if (mappedType != 0) {
                Printv(argList, mappedType, NIL);
            } else {
                Printf(stderr, "Can't convert argument type %s in function %s\n", oldType, wrapName);
                Printv(argList, oldType, NIL);
            }

            Printf(argList, " arg%i", i + 1);

            parm = nextSibling(parm);
        }

        Printf(wrapper->def, "%s %s(%s) {", "void", name, argList);
        emit_args(type, parms, wrapper);
        emit_action(node, wrapper);
        Wrapper_print(wrapper, d_source.get());

        return SWIG_OK;
    }

    int writeWrapperCtor(Node* node) {

        return SWIG_ERROR;
    }

    int writeWrapperDtor(Node* node) {
        return SWIG_ERROR;
    }

    void dumpNode(Node* node) {
            for (Iterator iter = First(node); iter.item; iter = Next(iter)) {
            if (DohIsString(iter.item)) {
                Printf(stderr, "'%s'\t'%s'\n", iter.key, iter.item);
            } else {
                Printf(stderr, "'%s'\t<%s>\n", iter.key, Getattr(iter.item, "nodeType"));
            }
        }
        Printf(stderr, "\n\n\n");
    }

    int NYI(Node* node) {
        Printf(stderr, "Not yet implemented: %s %s\n", Getattr(node, "nodeType"), Getattr(node, "name"));
        dumpNode(node);
        return SWIG_ERROR;
    }

    // attributes:

    /// Contains the implementation of the wrapped C API
    ScopedFile c_source;

    /// D source file that declares extern(C) functions for 
    /// the contents of c_source (above)
    ScopedFile d_import;

    /// Wraps the C API in a delicious, candy coated interface.
    ScopedFile d_source;

    /// Not used. (not useful?  Provide anyway for the sake of compatibility with other backends?)
#if 0
    ScopedString c_init;   // init code. (also goes into c_source, I think)
#endif

};   /* class D */

extern "C" Language *
swig_d(void) {
  return new D();
}

#if 0
const char *D::usage = reinterpret_cast<char*>("\
D Options (available with -d)\n\
     -package <name> - set name of the package to <name>\n\
\n");
#endif
